/* 
 This file has all of the ECM logic and supporting methods.
 
 The main entry point is the refreshDataFromApi() method. Once this method
 is called, it causes the following chained sequence of methods to be
 called:
 
 -> refreshDataFromApiResponseHandler()
	(if G_SKILL_MAP_DEFINED)	-> skillListResponseHandler()
									-> skillInTrainingResponseHandler()
									-> skillQueueResponseHandler()
									
	(if !G_SKILL_MAP_DEFINED)	-> skillInTrainingResponseHandler()
									-> skillQueueResponseHandler()

 */

// Setting this to true will cause the widget to load global preferences
// instead of instance preferences
var G_DEBUG = false;

// Widget version, this is compared to the current version on the website
var G_VERSION = "0.1.2";

// HTTP object for fetching data from the Eve API
// Note: concurrent request will break the methods that call this, so
// all request have to be chained sequentially
var G_HTTP;

// Skill map - holds skill ID to skill description mappings
// This is fetched the first time a skill needs to be looked up
// and is controlled by the boolean variable below
var G_SKILL_MAP = {};
var G_SKILL_MAP_DEFINED = false;

// Character's known skills and skill queue
var G_CHAR_SKILLS = {};
var G_CHAR_QUEUE = new Array();

// Current skill related information
var G_TQ_TIME;
var G_SKILL_FINISHES_TIME;
var G_SKILL_START_TIME;
var G_SKILL_ID;

// Timers
var G_TIMER_RUNNING = false;	// applies to the skill timer only
var G_SKILL_TIMER;	// set to fire when the current skill finishes
var G_HOURLY_TIMER;	// set to fire every hour after the widget is started

// View display flags
var G_BACK_SHOWN = false;
var G_ERROR_SHOWN = false;

//
// This class is used to render the skill queue, it feeds off the array
// containing the calculated skill data.
//
var G_SKILL_QUEUE_DATA = new SkillQueue();
function SkillQueue()
{
    this.numberOfRows = function()
	{
		return G_CHAR_QUEUE.length;
	}

    this.prepareRow = function(rowElement, rowIndex, templateElements)
	{
		if (templateElements.rowLabel)
		{
			templateElements.rowLabel.innerHTML = G_CHAR_QUEUE[rowIndex][0];
			templateElements.rowLabel2.innerHTML = G_CHAR_QUEUE[rowIndex][1];
		}
	}
}
	

//
// Retrieves a preference and sets the input field's value to the
// value stored for the preference key.
//
function loadPreferenceIntoInput(prefKey, inputField)
{
    var prefValue = loadSetting(prefKey);
    
    if (!(prefValue && prefValue.length > 0))
    {
        prefValue = "*Not Set*";
    }
    
    inputField.value = prefValue;
}

//
// Retrieves a preference and sets the text element value to the
// value stored for the preference key.
//
function loadPreferenceIntoText(prefKey, textElement)
{
    var prefValue = loadSetting(prefKey);
    
    if (!(prefValue && prefValue.length > 0))
    {
        prefValue = "*Not Set*";
    }
    
    textElement.innerText = prefValue;
}

//
// Saves a preference from an input field to the preference key
//
function savePreferenceFromInput(prefKey, inputField)
{
    var prefValue = inputField.value;
    saveSetting(prefKey, prefValue);
}

//
// Calls the Eve API using a POST HTTP method, the response is handled by the
// 'responseHandler' method provided. This is an async method.
//
function getApiData(method, userId, apiKey, charId, responseHandler) 
{
	// Check for internet connectivity
	var online = window.navigator.onLine;
	if (!online) {
		handleError("We are offline!");
		return;
	}

	// build the paramer string to pass to the API
    var params = "userID=" + userId + "&apiKey=" + apiKey;
	if (charId != null)
	{
		params = params + "&characterID=" + charId;
	}
	
	G_HTTP = new XMLHttpRequest();
    
    //req.onreadystatechange = processReqChange;
    G_HTTP.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
    G_HTTP.setRequestHeader("Content-length", params.length);
    G_HTTP.setRequestHeader("Connection", "close");
    
    G_HTTP.onreadystatechange=responseHandler;    
    G_HTTP.open("POST", "https://api.eveonline.com" + method, true);
	G_HTTP.send(params);
}

//
// Calls the Eve API to get a character list. The response is handled by
// the 'getCharactersFromApiResponseHandler' method.
//
function getCharactersFromApi() 
{
    var userID = loadSetting("userId");
    var apiKey = loadSetting("apiKey");
	
	// TODO: Check that we have all of the above variable set, error if not
		
    var data = getApiData("/account/Characters.xml.aspx", userID, apiKey, null, getCharactersFromApiResponseHandler);
}

//
// Processes the response from the Eve API for a character list and
// populates the preferences character selection drop down box.
//
function getCharactersFromApiResponseHandler()
{
	if (G_HTTP.readyState == 4)
	{
		var xmlDoc = G_HTTP.responseXML;
		if (xmlDoc == null)
		{
			handleError("Could not retrieve data while looking up characters.");
			return;
		}
			
		var chars = xmlDoc.getElementsByTagName("rowset")[0].getElementsByTagName("row");
		var numChars = chars.length;		
		var charList = new Array();
		charList.push(["", ""]);
			
		for (i = 0; i < numChars; i++)
		{ 
			var charName = chars[i].getAttribute("name");
			var charId = chars[i].getAttribute("characterID");
			charList.push([charName, charId]);
		}
		
		// Add all characters from API to the list
		var popupToChange = document.getElementById("charactersList");
		popupToChange.object.setOptions(charList);
		popupToChange.object.setEnabled(true);
	}
}

//
// TODO: Document
//
function handleError(message)
{
	console.log(message);
	errorText.innerText = message;
	
	if (!G_BACK_SHOWN)
	{
		frontErrorText.innerHTML = message;
		statusIndicator.innerText = "Error";
		
		newTransition = new Transition(Transition.CUBE_TYPE,1,Transition.EASE_TIMING);
		document.getElementById('stackLayout').object.setCurrentViewWithTransition('error',newTransition,false);

		G_ERROR_SHOWN = true;
	}
}

function refreshDataFromApiHourly()
{
	setTimeout("refreshDataFromApiHourly();", 3600000);
	nextRefreshValue.innerText = new Date((new Date()).getTime() + 3600000);
	refreshDataFromApi();
}

//
// Calls the Eve API to get the character sheet and current skill in training.
// The response is handled by the 'refreshDataFromApiResponseHandler' method.
function refreshDataFromApi()
{
	stopSkillTimer();
	
	statusIndicator.innerText = "Fetching character data...";

	var userID = loadSetting("userId");
    var apiKey = loadSetting("apiKey");
	var charId = loadSetting("charId");
	
	// TODO: Check that we have all of the above variable set, error if not
	
    getApiData("/char/CharacterSheet.xml.aspx", userID, apiKey, charId, refreshDataFromApiResponseHandler);
}

//
// TODO: Document
//
function refreshDataFromApiResponseHandler()
{
	if (G_HTTP.readyState == 4)
	{
		var xmlDoc = G_HTTP.responseXML;
		if (xmlDoc == null)
		{
			handleError("Could not retrieve data while looking up the character sheet.");
			return;
		}
		
		// Check for API errors i.e. backend being disabled
		if (checkApiError(xmlDoc)) {
			return;
		}
		
		characterLabel.innerText = getNodeValue(xmlDoc, "name");
		corporationLabel.innerText = getNodeValue(xmlDoc, "corporationName");
		balanceValue.innerText = formatAsCurrency(getNodeValue(xmlDoc, "balance"));
		characterImage.innerHTML = 
			"<img src='http://image.eveonline.com/Character/" + loadSetting("charId") + 
			"_64.jpg' width='64' height='64' border='0' />";
			
		characterImage_bk.innerHTML = 
			"<img src='http://image.eveonline.com/Character/" + loadSetting("charId") + 
			"_256.jpg' width='268' height='256' border='0' style='Filter: Alpha(Opacity=100, FinishOpacity=0, Style=2, StartX=20, StartY=40, FinishX=0, FinishY=0)'/>";
			
		var skills = getRowSetByName(xmlDoc.getElementsByTagName("rowset"), "skills").getElementsByTagName("row");
		var totalSp = 0;
		for (i = 0; i < skills.length; i++)
		{
			var skillPoints = skills[i].getAttribute("skillpoints");
			var skillId = skills[i].getAttribute("typeID");
			
			totalSp += (skillPoints * 1);
			G_CHAR_SKILLS[skillId] = skillPoints;
			
			//requiredAttributes
			//primaryAttribute
			//secondaryAttribute
			
			//memory
			//charisma
			
			//effectiveAttribute = (baseAttribute + basicLearningSkill + advancedLearningSkill + Implant) * (1 + (0.02 * LearningLevel))
		}
		totalSpValue.innerText = formatAsCurrency(totalSp);
		var cloneSp = getNodeValue(xmlDoc, "cloneSkillPoints")

		if (totalSp > cloneSp)
		{
			cloneSpValue.innerHTML = "<font color='Red'>" + formatAsCurrency(cloneSp) + "</font>";
		}
		else
		{
			cloneSpValue.innerHTML = "<font color='Green'>" + formatAsCurrency(cloneSp) + "</font>";
		}

		
		if (!G_SKILL_MAP_DEFINED)
		{
			statusIndicator.innerText = "Fetching skill tree...";
			
			getApiData("/eve/SkillTree.xml.aspx", null, null, null, skillListResponseHandler);
		}
		else
		{
			statusIndicator.innerText = "Fetching training data...";
			
			var userID = loadSetting("userId");
			var apiKey = loadSetting("apiKey");
			var charId = loadSetting("charId");
			getApiData("/char/SkillInTraining.xml.aspx", userID, apiKey, charId, skillInTrainingResponseHandler);
		}
	}
}

function skillInTrainingResponseHandler()
{
	if (G_HTTP.readyState == 4)
	{
		var xmlDoc = G_HTTP.responseXML;
		if (xmlDoc == null)
		{
			handleError("Could not retrieve data while looking up the current skill in training.");
			return;
		}
		
		var isTraining = getNodeValue(xmlDoc, "skillInTraining");
		if (isTraining == "0")
		{
			skillInTraining.innerHTML = "<font color='Red'>* No Skill Training *</font>";
			finishesOnValue.innerText = "N/A";
			timeRemainingValue.innerText = "N/A";
			document.getElementById("trainingProgress").object.setValue(0);
			trainingPercent.innerText = "N/A";
			trainingRateValue.innerText = "N/A";
			statusIndicator.innerText = "";
			queueTotalTimeValue.innerHTML = "0d 0m 0s";

			G_CHAR_QUEUE = new Array();
			document.getElementById("list").object.reloadData();
			
			return;
		}
		
		G_TQ_TIME = parseEveDate(getNodeValue(xmlDoc, "currentTQTime"));
		G_SKILL_FINISHES_TIME = parseEveDate(getNodeValue(xmlDoc, "trainingEndTime"));
		G_SKILL_START_TIME = parseEveDate(getNodeValue(xmlDoc, "trainingStartTime"));
		startSkillTimer();
		
		G_SKILL_ID = getNodeValue(xmlDoc, "trainingTypeID");
		var skillLevel = getNodeValue(xmlDoc, "trainingToLevel");
		var trainingStartSP = getNodeValue(xmlDoc, "trainingStartSP");
		var trainingDestinationSP = getNodeValue(xmlDoc, "trainingDestinationSP");
		var trainingRate = Math.floor((trainingDestinationSP - trainingStartSP) / ((G_SKILL_FINISHES_TIME - G_SKILL_START_TIME) / 3600000));

		skillInTraining.innerText = G_SKILL_MAP[G_SKILL_ID] + " " + skillLevel;
		trainingRateValue.innerText = trainingRate + " SP/h";
		
		statusIndicator.innerText = "Fetching queue data...";
		var userID = loadSetting("userId");
		var apiKey = loadSetting("apiKey");
		var charId = loadSetting("charId");
		getApiData("/char/SkillQueue.xml.aspx", userID, apiKey, charId, skillQueueResponseHandler);
	}
}

function skillQueueResponseHandler()
{
	if (G_HTTP.readyState == 4)
	{
		var xmlDoc = G_HTTP.responseXML;
		if (xmlDoc == null)
		{
			handleError("Could not retrieve data while looking up the training queue.");
			return;
		}
		
		G_CHAR_QUEUE = new Array();
		var queueStart = null;
		var queueFinish = null;
		
		var skills = getRowSetByName(xmlDoc.getElementsByTagName("rowset"), "skillqueue").getElementsByTagName("row");
		for (i = 0; i < skills.length; i++)
		{
			var skillId = skills[i].getAttribute("typeID");
			var skillLevel = skills[i].getAttribute("level");
			
			var finish = parseEveDate(skills[i].getAttribute("endTime"));
			var start = parseEveDate(skills[i].getAttribute("startTime"));
			
			// adjust for currently training skill remaining time
			if (skillId == G_SKILL_ID)
			{
				start = new Date(finish - (finish - G_TQ_TIME));
			}
			
			if (i == 0)
			{
				queueStart = start;
			}
			queueFinish = finish;
			
			var rowData = {};
			rowData[0] = "" + G_SKILL_MAP[skillId] + " " + skillLevel;
			rowData[1] = getTimeString(finish - start);
			G_CHAR_QUEUE.push(rowData);
		}
		
		queueTotalTimeValue.innerHTML = getTimeString(queueFinish - queueStart);
		
		document.getElementById("list").object.reloadData();
		
		statusIndicator.innerText = "";
		
		// if the error view is shown and we get all data ok, transition to the character sheet
		if (G_ERROR_SHOWN)
		{
			transitionView('stackLayout', 'characterSheet');
			G_ERROR_SHOWN = false;
		}
	}
}

function startSkillTimer()
{
	if (!G_TIMER_RUNNING)
	{
		G_SKILL_TIMER = setTimeout("updateTqTime();", 1000);
		G_TIMER_RUNNING = true;
	}
}

function stopSkillTimer()
{
	G_TIMER_RUNNING = false;
	clearTimeout(G_SKILL_TIMER);
}

function updateTqTime()
{
	G_TQ_TIME = new Date(G_TQ_TIME.getTime() + 1000);
	populateCountDowns();

	if (G_TIMER_RUNNING)
	{
		G_SKILL_TIMER = setTimeout("updateTqTime();", 1000);
	}
}

function populateCountDowns()
{
	var needUpdate = false;
	var trainingDiff = G_SKILL_FINISHES_TIME - G_TQ_TIME;
	if (trainingDiff < 1000)
	{
		needUpdate = true;
	}

	var percComplete = Math.floor((1 - (trainingDiff / (G_SKILL_FINISHES_TIME - G_SKILL_START_TIME))) * 100);
		
	timeRemainingValue.innerHTML = getTimeString(trainingDiff);
	finishesOnValue.innerText = G_SKILL_FINISHES_TIME;
	
	var curentProgress = document.getElementById("trainingProgress").object.getValue();
	if (curentProgress != percComplete)
	{
		document.getElementById("trainingProgress").object.setValue(percComplete);
		trainingPercent.innerText = percComplete + "%";
	}

	if (needUpdate)
	{
		refreshDataFromApi();
	}
}

function getTimeString(timeStamp)
{
	var daysLeft = Math.floor(timeStamp / 86400000);
	timeStamp -= daysLeft * 86400000;
	var hoursLeft = Math.floor(timeStamp / 3600000);
	timeStamp -= hoursLeft * 3600000;
	var minutesLeft = Math.floor(timeStamp / 60000);
	timeStamp -= minutesLeft * 60000;
	var secondsLeft = Math.floor(timeStamp / 1000);
	
	if (daysLeft == 0)
	{
		return "<font color='Red'>" + daysLeft + "d " + hoursLeft + "h " + minutesLeft + "m " + secondsLeft + "s</font>";
	}
	else
	{			
		return daysLeft + "d " + hoursLeft + "h " + minutesLeft + "m " + secondsLeft + "s";
	}
}

function parseEveDate(eveDate)
{
	//2008-08-17 06:43:00
	var x = eveDate.split(" ");
	var datePart = x[0];
	var timePart = x[1];
	
	var y = datePart.split("-");
	var z = timePart.split(":");
	
	//new Date(year, month, day, hours, minutes, seconds, milliseconds);
	return new Date(y[0], y[1] - 1, y[2], z[0], z[1], z[2]);
}

//
// Searches for a rowset with a given name from an array of rowsets.
//
function getRowSetByName(rowSets, rowSetName)
{
	for (i = 0; i < rowSets.length; i++)
	{
		var currentRowSetName = rowSets[i].getAttribute("name");
		if (currentRowSetName == rowSetName)
		{
			return rowSets[i];
		}
	}
	
	return null;
}


function formatAsCurrency(nStr)
{
	nStr += '';
	x = nStr.split('.');
	x1 = x[0];
	var rgx = /(\d+)(\d{3})/;
	while (rgx.test(x1)) {
		x1 = x1.replace(rgx, '$1' + ',' + '$2');
	}
	return x1;
}

function skillListResponseHandler()
{
	if (G_HTTP.readyState == 4)
	{
		var xmlDoc = G_HTTP.responseXML;
		if (xmlDoc == null)
		{
			handleError("Could not retrieve data while looking up the skill list.");
			return;
		}
		
		var skills = xmlDoc.getElementsByTagName("rowset")[0].getElementsByTagName("row");
		for (i = 0; i < skills.length; i++)
		{
			var skillName = skills[i].getAttribute("typeName");
			
			if (skillName != null)
			{
				var skillId = skills[i].getAttribute("typeID");
				G_SKILL_MAP[skillId] = skillName;
			}
		}
		
		G_SKILL_MAP_DEFINED = true;
		
		statusIndicator.innerText = "Fetching training data...";
		
		var userID = loadSetting("userId");
		var apiKey = loadSetting("apiKey");
		var charId = loadSetting("charId");
		getApiData("/char/SkillInTraining.xml.aspx", userID, apiKey, charId, skillInTrainingResponseHandler);
	}
}

function saveSetting(key, value)
{
	if (G_DEBUG)
	{
		widget.setPreferenceForKey(value, key);
	}
	else
	{
		widget.setPreferenceForKey(value, widget.identifier + "-" + key);
	}
}

function loadSetting(key)
{
	if (G_DEBUG)
	{
		return widget.preferenceForKey(key);
	}
	else
	{
		return widget.preferenceForKey(widget.identifier + "-" + key);
	}
}

function getNodeValue(xmlDocument, elementTagName)
{
	return xmlDocument.getElementsByTagName(elementTagName)[0].childNodes[0].nodeValue;
}

//
// Checks the document for an 'error' tag. Displays the error message if
// the tag is found and returns true, othewise returns false.
//
function checkApiError(xmlDocument)
{
	var error = xmlDocument.getElementsByTagName("error")[0];
	if (error != null)
	{
		var errorText = error.childNodes[0].nodeValue;
		handleError(errorText);
		return true;
	}
	
	return false;
}

function checkForUpdates()
{
	statusIndicator.innerText = "Checking for updates...";

	// Check for internet connectivity
	var online = window.navigator.onLine;
	if (!online) {
		handleError("We are offline!");
		return;
	}
	
	G_HTTP = new XMLHttpRequest();
    G_HTTP.setRequestHeader("Connection", "close");
    G_HTTP.onreadystatechange = checkForUpdatesResponseHandler;    
    G_HTTP.open("GET", "http://www.colorsin.com/ecm.version", true);
	G_HTTP.send(null);
}

function checkForUpdatesResponseHandler()
{
	if (G_HTTP.readyState == 4)
	{
		var txtDoc = G_HTTP.responseText;
		if (txtDoc == null || txtDoc == "")
		{
			handleError("Could not retrieve data while checking for updates.");
			return;
		}
		
		if (txtDoc.indexOf(G_VERSION) == -1) {
			handleError("You are not using the most up to date version, please download a new version from<br><br><a href='http://www.colorsin.com/ecm.html'>http://www.colorsin.com/ecm.html</a>");
		}
		else
		{
			refreshDataFromApi();
		}
	}
}
